#import "XeeBitmapImage.h"



@implementation XeeBitmapImage

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes
{
	if(self=[super initWithFile:name firstBlock:block attributes:attributes])
	{
		type=0;
		bytesperrow=0;
		pixelsize=0;
		premultiplied=NO;

		data=NULL;
		freedata=NO;

		textarget=texintformat=texformat=textype=0;

		gw=NULL;
		imagerep=nil;

		tiles=nil;
		needsupdate=NO;
		context=nil;
	}

	return self;
}

-(id)initWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight
{
	if(self=[self initWithFile:nil firstBlock:nil attributes:nil])
	{
		if([self allocBitmapWithType:pixeltype width:framewidth height:frameheight])
		{
			return self;
		}

		[self release];
	}

	return nil;
}

-(id)initWithData:(void *)pixeldata type:(int)pixeltype width:(int)framewidth height:(int)frameheight bytesPerRow:(int)bprow
{
	if(self=[self initWithFile:nil firstBlock:nil attributes:nil])
	{
		[self setData:pixeldata type:pixeltype width:framewidth height:frameheight bytesPerRow:bprow];

		return self;
	}
	return nil;
}

-(id)initWithImageRep:(NSBitmapImageRep *)rep
{
	if(self=[self initWithFile:nil firstBlock:nil attributes:nil])
	{
		if(![rep isPlanar])
		{
			int bitmapformat=[rep respondsToSelector:@selector(bitmapFormat)]?[rep bitmapFormat]:0;
			int pixeltype;

			switch([rep bitsPerPixel])
			{
				case 8:
					pixeltype=XeeBitmapTypeLuma8;
					[self setDepthGrey:256];
				break;

				case 24:
					pixeltype=XeeBitmapTypeRGB24;
					[self setDepthRGB:24];
				break;

				case 32:
					if(bitmapformat&NSAlphaFirstBitmapFormat) pixeltype=XeeBitmapTypeARGB32;
					else pixeltype=XeeBitmapTypeRGBA32;
					if([rep samplesPerPixel]==4) [self setDepthRGBA:32];
					else [self setDepthRGB:24];
				break;

				default: pixeltype=0; break;
			}

			if(pixeltype)
			{
				[self setData:[rep bitmapData] type:pixeltype width:[rep pixelsWide] height:[rep pixelsHigh] bytesPerRow:[rep bytesPerRow]];

				if([rep samplesPerPixel]==4)
				{
					[self setTransparent:YES];
					[self setBackgroundColor:[NSColor whiteColor]];
				}

				if(!(bitmapformat&NSAlphaNonpremultipliedBitmapFormat)) [self setPremultiplied:YES];

				imagerep=[rep retain];

				return self;
			}
		}
		[self release];
	}
	return nil;
}

-(void)dealloc
{
	[context makeCurrentContext];
	[tiles release];
	[context release];

	if(freedata) free(data);

	if(gw) DisposeGWorld(gw);
	[imagerep release];

	[super dealloc];
}

-(void)setData:(void *)pixeldata type:(int)pixeltype width:(int)framewidth height:(int)frameheight bytesPerRow:(int)bprow
{
	width=framewidth;
	height=frameheight;
	data=pixeldata;
	type=pixeltype;
	bytesperrow=bprow;

	switch(type)
	{
		case XeeBitmapTypeLuma8:
			texintformat=GL_LUMINANCE8;
			texformat=GL_LUMINANCE;
			textype=GL_UNSIGNED_BYTE;
			pixelsize=1;
		break;

		case XeeBitmapTypeYUV422:
			texintformat=GL_RGB8;
			texformat=GL_YCBCR_422_APPLE;
			textype=GL_UNSIGNED_SHORT_8_8_REV_APPLE;
			pixelsize=2;
		break;

		case XeeBitmapTypeRGB24:
			texintformat=GL_RGB;
			texformat=GL_RGB;
			textype=GL_UNSIGNED_BYTE;
			pixelsize=3;
		break;

		case XeeBitmapTypeARGB32:
			texintformat=GL_RGBA8;
			texformat=GL_BGRA;
			textype=GL_UNSIGNED_INT_8_8_8_8_REV;
			pixelsize=4;
		break;

		case XeeBitmapTypeRGBA32:
			texintformat=GL_RGBA8;
			texformat=GL_RGBA;
			textype=GL_UNSIGNED_BYTE;
			pixelsize=4;
		break;
	}
}

-(BOOL)allocBitmapWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight
{
	if(data) return NO; // maybe dealloc?

	int bprow;
	switch(pixeltype)
	{
		case XeeBitmapTypeLuma8: bprow=framewidth; break;
		case XeeBitmapTypeYUV422: bprow=(2*framewidth+3)&~3; break;
		case XeeBitmapTypeRGB24: bprow=(3*framewidth+3)&~3; break;
		case XeeBitmapTypeARGB32: bprow=4*framewidth; break;
		case XeeBitmapTypeRGBA32: bprow=4*framewidth; break;
	}

	void *pixeldata=malloc(bprow*frameheight);

	if(pixeldata)
	{
		[self setData:pixeldata type:pixeltype width:framewidth height:frameheight bytesPerRow:bprow];

		freedata=YES;

		return YES;
	}

	return NO;
}

-(void)setPremultiplied:(BOOL)pre { premultiplied=pre; }

-(NSBitmapImageRep *)imageRep
{
	if(!imagerep)
	{
		int bitsperpixel,samplesperpixel;
		int bitmapformat=0;
		BOOL hasalpha;

		switch(type)
		{
			case XeeBitmapTypeLuma8:
				bitsperpixel=8;
				samplesperpixel=1;
				hasalpha=NO;
			break;

			case XeeBitmapTypeRGB24:
				bitsperpixel=24;
				samplesperpixel=3;
				hasalpha=NO;
			break;

			case XeeBitmapTypeRGBA32:
				bitsperpixel=32;
				samplesperpixel=4;
				hasalpha=YES;
			break;

			case XeeBitmapTypeARGB32:
				bitsperpixel=32;
				samplesperpixel=4;
				bitmapformat|=NSAlphaFirstBitmapFormat;
				hasalpha=YES;
			break;

			case XeeBitmapTypeYUV422: // always needs to be converted
				imagerep=[[self makeImageRep] retain];
				return imagerep;
			break;
		}

		if(hasalpha&&!premultiplied) bitmapformat|=NSAlphaNonpremultipliedBitmapFormat;

		if([NSBitmapImageRep instancesRespondToSelector:@selector(initWithBitmapDataPlanes:pixelsWide:pixelsHigh:bitsPerSample:samplesPerPixel:hasAlpha:isPlanar:colorSpaceName:bitmapFormat:bytesPerRow:bitsPerPixel:)])
		{
			imagerep=[[NSBitmapImageRep alloc]
			initWithBitmapDataPlanes:(unsigned char **)&data pixelsWide:width pixelsHigh:height bitsPerSample:8
			samplesPerPixel:samplesperpixel hasAlpha:hasalpha isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace
			bitmapFormat:bitmapformat bytesPerRow:bytesperrow bitsPerPixel:bitsperpixel];
		}
		else
		{
			if(bitmapformat) imagerep=[[self makeImageRep] retain];
			else imagerep=[[NSBitmapImageRep alloc]
			initWithBitmapDataPlanes:(unsigned char **)&data pixelsWide:width pixelsHigh:height bitsPerSample:8
			samplesPerPixel:samplesperpixel hasAlpha:hasalpha isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace
			bytesPerRow:bytesperrow bitsPerPixel:bitsperpixel];
		}
	}

	return imagerep;
}

-(NSBitmapImageRep *)makeImageRep
{
	NSBitmapImageRep *rep=nil;

	switch(type)
	{
		case XeeBitmapTypeLuma8:
		case XeeBitmapTypeRGB24:
		{
			int samples;
			if(type==XeeBitmapTypeLuma8) samples=1;
			else samples=3;

			rep=[[[NSBitmapImageRep alloc]
			initWithBitmapDataPlanes:NULL pixelsWide:width pixelsHigh:height bitsPerSample:8
			samplesPerPixel:samples hasAlpha:NO isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace
			bytesPerRow:0 bitsPerPixel:samples*8] autorelease];

			unsigned char *destdata=[rep bitmapData];
			int destbytesperrow=[rep bytesPerRow];

			for(int y=0;y<height;y++)
			{
				unsigned char *srcrow=data+y*bytesperrow;
				unsigned char *destrow=destdata+y*destbytesperrow;
				memcpy(destrow,srcrow,width*samples);
			}
		}
		break;

		case XeeBitmapTypeARGB32:
		case XeeBitmapTypeRGBA32:
		{
			rep=[[[NSBitmapImageRep alloc]
			initWithBitmapDataPlanes:NULL pixelsWide:width pixelsHigh:height bitsPerSample:8
			samplesPerPixel:4 hasAlpha:YES isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace
			bytesPerRow:0 bitsPerPixel:32] autorelease];

			unsigned char *destdata=[rep bitmapData];
			int destbytesperrow=[rep bytesPerRow];

			for(int y=0;y<height;y++)
			{
				unsigned long *srcrow=(unsigned long *)(data+y*bytesperrow);
				unsigned long *destrow=(unsigned long *)(destdata+y*destbytesperrow);

				for(int x=0;x<width;x++)
				{
					unsigned long val=*srcrow++;
					unsigned char a,r,g,b;

					if(type==XeeBitmapTypeRGBA32) { r=val>>24; g=val>>16; b=val>>8; a=val; }
					else { a=val>>24; r=val>>16; g=val>>8; b=val; }

					if(!premultiplied) { r=(r*a)/255; g=(g*a)/255; b=(b*a)/255; }

					*destrow++=(r<<24)|(g<<16)|(b<<8)|a;
				}
			}
		}
		break;

		case XeeBitmapTypeYUV422: // cheat and load from disk
		{
			NSArray *reps=[NSBitmapImageRep imageRepsWithData:[NSData dataWithContentsOfFile:filename]];

			if(!reps||![reps count]) return nil;

			NSEnumerator *enumerator=[reps objectEnumerator];

			while(rep=[enumerator nextObject])
			{
				if([rep pixelsWide]==width&&[rep pixelsHigh]==height) return rep;
			}

			return [reps objectAtIndex:0];
		}
		break;
	}

	return rep;
}

-(GWorldPtr)gworld
{
	if(!gw)
	{
		OSType pixelformat;

		switch(type)
		{
			case XeeBitmapTypeLuma8: pixelformat=k8IndexedGrayPixelFormat; break;
			case XeeBitmapTypeYUV422: pixelformat=k2vuyPixelFormat; break;
			case XeeBitmapTypeRGB24: pixelformat=k24RGBPixelFormat; break;
			case XeeBitmapTypeARGB32: pixelformat=k32ARGBPixelFormat; break;
			case XeeBitmapTypeRGBA32: pixelformat=k32RGBAPixelFormat; break; // doesn't work!
			default: pixelformat=0; break;
		}

		if(pixelformat)
		{
			Rect rect;
			SetRect(&rect,0,0,width,height);
			QTNewGWorldFromPtr(&gw,pixelformat,&rect,NULL,NULL,0,data,bytesperrow);
		}
	}

	return gw;
}




-(void)triggerUpdateAction
{
	needsupdate=YES;
	[super triggerUpdateAction];
}

-(void)draw:(NSPoint)point size:(NSSize)size bounds:(NSRect)bounds
{
	if(!data) return;

	if(!tiles)
	{
		context=[[NSOpenGLContext currentContext] retain];

		if(![[NSUserDefaults standardUserDefaults] boolForKey:@"force2D"]&&
		gluCheckExtension((unsigned char *)"GL_EXT_texture_rectangle",glGetString(GL_EXTENSIONS)))
		[self allocTexturesRect];
		else
		[self allocTextures2D];

		needsupdate=YES;
	}

	if(needsupdate)
	{
		needsupdate=NO;
		[self loadTextures];
	}

	glPushMatrix();

	float x_scale=size.width/(float)width;
	float y_scale=size.height/(float)height;

	NSRect transbounds=NSMakeRect(
		(bounds.origin.x-point.x)/x_scale,
		(bounds.origin.y-point.y)/y_scale,
		bounds.size.width/x_scale,
		bounds.size.height/y_scale);

	glTranslatef(point.x,point.y,0);
	glScalef(x_scale,y_scale,1);

	glEnable(textarget);
	glEnable(GL_BLEND);

	if(premultiplied) glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
	else glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	NSEnumerator *enumerator=[tiles objectEnumerator];
	XeeBitmapTile *tile;

	while(tile=[enumerator nextObject])
	{
		[tile drawWithTarget:textarget bounds:transbounds]; // needs a check
	}

	glPopMatrix();
}

-(void)allocTexturesRect
{
	textarget=GL_TEXTURE_RECTANGLE_EXT;

	GLint maxtilesize;
	glGetIntegerv(GL_MAX_RECTANGLE_TEXTURE_SIZE_EXT,&maxtilesize);
	if(maxtilesize>512) maxtilesize=512;

	int cols=(width+maxtilesize-1)/maxtilesize;
	int rows=(height+maxtilesize-1)/maxtilesize;

	tiles=[[NSMutableArray alloc] initWithCapacity:rows*cols];

	if(tiles)
	{
		for(int row=0;row<rows;row++)
		{
			int tile_h=(row==rows-1)?(height-(rows-1)*maxtilesize):maxtilesize;

			for(int col=0;col<cols;col++)
			{
				int tile_w=(col==cols-1)?(width-(cols-1)*maxtilesize):maxtilesize;

				XeeBitmapTile *tile=[[[XeeBitmapTile alloc] initWithTarget:textarget x:col*maxtilesize y:row*maxtilesize
				width:tile_w height:tile_h texwidth:tile_w texheight:tile_h] autorelease];
				if(tile) [tiles addObject:tile];
			}
		}
	}
}

-(void)allocTextures2D
{
	textarget=GL_TEXTURE_2D;

	GLint maxtilesize;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE,&maxtilesize);
	if(maxtilesize>512) maxtilesize=512;

	int colsize,rowsize;
	int x,y;

	tiles=[[NSMutableArray alloc] initWithCapacity:32];

	if(tiles)
	{
		rowsize=maxtilesize;
		y=0;
		while(y<height)
		{
			while(y+rowsize>height) rowsize/=2;

			colsize=maxtilesize;
			x=0;
			while(x<width)
			{
				while(x+colsize>width) colsize/=2;

				XeeBitmapTile *tile=[[[XeeBitmapTile alloc] initWithTarget:textarget
				x:x y:y width:colsize height:rowsize texwidth:1 texheight:1] autorelease];
				if(tile) [tiles addObject:tile];

				x+=colsize;
			}

			y+=rowsize;
		}
	}
}

-(void)loadTextures
{
	int align;
	if((bytesperrow&7)==0) align=8;
	else if((bytesperrow&3)==0) align=4;
	else if((bytesperrow&1)==0) align=2;
	else align=1;

	glPixelStorei(GL_UNPACK_ROW_LENGTH,bytesperrow/pixelsize);
	glPixelStorei(GL_UNPACK_ALIGNMENT,align);

	if(textarget==GL_TEXTURE_2D&&(texformat==GL_LUMINANCE||texformat==GL_YCBCR_422_APPLE)) // workaround for buggy Rage128 drivers
//	if(textarget==GL_TEXTURE_2D&&texformat==GL_LUMINANCE) // workaround for buggy Rage128 drivers
	glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,GL_FALSE);
	else glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,GL_TRUE);

	NSEnumerator *enumerator=[tiles objectEnumerator];
	XeeBitmapTile *tile;

	while(tile=[enumerator nextObject])
	{
		[tile loadWithTarget:textarget internalFormat:texintformat format:texformat type:textype data:data];
	}
}



-(void)fixYUVGamma
{
	if(type!=XeeBitmapTypeYUV422) return;

	static const unsigned char gammatable[256]=
	{
		0x00,0x00,0x01,0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x05,0x05,0x06,0x07,0x07,0x08,
		0x09,0x09,0x0a,0x0b,0x0b,0x0c,0x0d,0x0d,0x0e,0x0f,0x10,0x10,0x11,0x12,0x13,0x13,
		0x14,0x15,0x16,0x17,0x17,0x18,0x19,0x1a,0x1b,0x1b,0x1c,0x1d,0x1e,0x1f,0x1f,0x20,
		0x21,0x22,0x23,0x24,0x25,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2c,0x2d,0x2e,
		0x2f,0x30,0x31,0x32,0x33,0x34,0x35,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,
		0x3e,0x3f,0x40,0x41,0x42,0x43,0x44,0x45,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,
		0x4d,0x4e,0x4f,0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x5b,0x5c,
		0x5d,0x5e,0x5f,0x60,0x61,0x62,0x63,0x64,0x65,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,
		0x6e,0x6f,0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7b,0x7c,0x7d,0x7e,
		0x7f,0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
		0x90,0x91,0x92,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9e,0x9f,0xa0,0xa1,
		0xa2,0xa3,0xa4,0xa5,0xa7,0xa8,0xa9,0xaa,0xab,0xac,0xad,0xaf,0xb0,0xb1,0xb2,0xb3,
		0xb4,0xb5,0xb7,0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbf,0xc0,0xc1,0xc2,0xc3,0xc4,0xc6,
		0xc7,0xc8,0xc9,0xca,0xcb,0xcd,0xce,0xcf,0xd0,0xd1,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,
		0xda,0xdb,0xdc,0xdd,0xde,0xe0,0xe1,0xe2,0xe3,0xe4,0xe6,0xe7,0xe8,0xe9,0xea,0xec,
		0xed,0xee,0xef,0xf0,0xf2,0xf3,0xf4,0xf5,0xf6,0xf8,0xf9,0xfa,0xfb,0xfd,0xfe,0xff,
	};

	int linewidth=4*((width+1)/2);

	for(int y=0;y<height;y++)
	{
		int n=width;
		unsigned char *ptr=(unsigned char *)data+y*linewidth+1;

		while(n--)
		{
			*ptr=gammatable[*ptr];
			ptr+=2;
		}
	}
}

-(void)invertChannels
{
	int totalbytes=bytesperrow*height;
	int totallongs=totalbytes/4;
	int extrabytes=totalbytes%4;

	unsigned long *ptr=(unsigned long *)data;

	while(totallongs--)
	{
		*ptr=~*ptr;
		ptr++;
	}

	unsigned char *cptr=(unsigned char *)ptr;

	while(extrabytes--)
	{
		*cptr=~*cptr;
		cptr++;
	}
}



-(int)frames { return 1; }

-(int)type { return type; }

-(int)bytesPerRow { return bytesperrow; }

-(void *)data { return data; }

@end



// XeeBitmapTile

@implementation XeeBitmapTile

-(id)initWithTarget:(GLuint)textarget x:(int)x0 y:(int)y0 width:(int)w height:(int)h texwidth:(float)tw texheight:(float)th
{
	if(self=[super init])
	{
		x=x0;
		y=y0;
		width=w;
		height=h;
		texwidth=tw;
		texheight=th;

		glGenTextures(1,&tex);
		if(tex)
		{
			glBindTexture(textarget,tex);

//			glTexParameteri(textarget,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
			glTexParameteri(textarget,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(textarget,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
			glTexParameteri(textarget,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
			glTexParameteri(textarget,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
			glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

			return self;
		}

		[self release];
	}

	return nil;
}

-(void)dealloc
{
	glDeleteTextures(1,&tex);

	[super dealloc];
}

-(void)loadWithTarget:(GLuint)textarget internalFormat:(GLuint)texintformat format:(GLuint)texformat type:(GLuint)textype data:(void *)data
{
	glBindTexture(textarget,tex);

	glPixelStorei(GL_UNPACK_SKIP_PIXELS,x);
	glPixelStorei(GL_UNPACK_SKIP_ROWS,y);

	int realwidth;

	if(texformat==GL_YCBCR_422_APPLE) // fudge odd-width YUV textures
	{
		realwidth=(width+1)&~1;
		if(textarget==GL_TEXTURE_2D&&width==1) texwidth=0.5;
	}
	else realwidth=width;

	glTexImage2D(textarget,0,texintformat,realwidth,height,0,texformat,textype,data);
}

-(void)drawWithTarget:(GLuint)textarget bounds:(NSRect)bounds
{
	if(!tex) return;
	if(!NSIntersectsRect(NSMakeRect(x,y,width,height),bounds)) return;

	glPushMatrix();
	glTranslatef(x,y,0);

	glBindTexture(textarget,tex);

	glBegin(GL_QUADS);

	glTexCoord2f(0,texheight);
	glVertex2i(0,height);

	glTexCoord2f(texwidth,texheight);
	glVertex2i(width,height);

	glTexCoord2f(texwidth,0);
	glVertex2i(width,0);

	glTexCoord2f(0,0);
	glVertex2i(0,0);


	glEnd();
		
	glPopMatrix();
}

@end
