#import <Cocoa/Cocoa.h>

#define XeeNoMode 0
#define XeeMoveMode 1
#define XeeCopyMode 2

@class XeeImage,XeeView,XeeStatusBar;

@interface XeeController:NSObject
{
	XeeImage *currimage;
	float zoom;

	int window_focus_x,window_focus_y;
	BOOL blocked;

	NSToolbar *toolbar;
	NSDictionary *toolbaritems;
	NSArray *toolbaridentifiers,*defaultidentifiers;

	IBOutlet NSWindow *window;
    IBOutlet XeeView *imageview;
    IBOutlet XeeStatusBar *statusbar;
}

-(id)init;
-(void)dealloc;
-(void)awakeFromNib;

-(void)windowWillClose:(NSNotification *)notification;
-(void)windowWillMiniaturize:(NSNotification *)notification;
-(void)windowDidMiniaturize:(NSNotification *)notification;

-(NSToolbarItem *)toolbar:(NSToolbar *)toolbar itemForItemIdentifier:(NSString *)identifier willBeInsertedIntoToolbar:(BOOL)flag;
-(NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar *)toolbar;
-(void)setupToolbarItems;
-(NSArray *)makeToolbarItems;
-(NSArray *)makeDefaultToolbarItemIdentifiers;

-(BOOL)validateMenuItem:(id <NSMenuItem>)item;
-(BOOL)validateAction:(SEL)action;

-(void)setupStatusBar;
-(void)updateStatusBar;
-(void)setStatusBarHidden:(BOOL)hidden;
-(BOOL)isStatusBarHidden;
-(IBAction)toggleStatusBar:(id)sender;

-(NSWindow *)window;

-(void)setImage:(XeeImage *)image;
-(void)imageUpdated:(XeeImage *)image;

-(void)setZoom:(float)newzoom;
-(void)setFrame:(int)frame;

-(void)updateWindowPosition;
-(void)setViewSize:(NSSize)size;
-(void)setStandardViewSize;
-(void)setResizeBlock:(BOOL)block;
-(void)setResizeBlockFromSender:(id)sender;
-(BOOL)isResizeBlocked;
-(NSSize)maxViewSize;
-(NSSize)minViewSize;
-(NSRect)availableScreenSpace;

-(IBAction)copy:(id)sender;
-(IBAction)saveImage:(id)sender;

-(IBAction)frameSkipNext:(id)obj;
-(IBAction)frameSkipPrev:(id)obj;
-(IBAction)toggleAnimation:(id)obj;

-(IBAction)zoomIn:(id)sender;
-(IBAction)zoomOut:(id)sender;
-(IBAction)zoomActual:(id)sender;
-(IBAction)zoomFit:(id)sender;

-(IBAction)fullScreen:(id)sender;

-(IBAction)closeWindowOrDrawer:(id)sender;

@end



@interface XeeWindow:NSWindow
{
}

@end



@interface NSToolbarItem (QuickConstructor)

+(NSToolbarItem *)itemWithIdentifier:(NSString *)ident label:(NSString *)label image:(NSString *)imagename action:(SEL)action;
+(NSToolbarItem *)itemWithIdentifier:(NSString *)ident label:(NSString *)label view:(NSView *)view;

@end



@interface XeeSegmentedItem:NSToolbarItem
{
	NSSegmentedControl *control;
	NSMenu *menu;
	SEL *actions;
}

-(id)initWithItemIdentifier:(NSString *)identifier label:(NSString *)label paletteLabel:(NSString *)pallabel segments:(int)segments;
-(void)dealloc;

-(void)setSegment:(int)segment label:(NSString *)label image:(NSImage *)image longLabel:(NSString *)longlabel width:(int)width action:(SEL)action;
-(void)setSegment:(int)segment label:(NSString *)label longLabel:(NSString *)longlabel action:(SEL)action;
-(void)setSegment:(int)segment imageName:(NSString *)imagename longLabel:(NSString *)longlabel action:(SEL)action;
-(void)setupView;

-(void)clicked:(id)sender;

+(XeeSegmentedItem *)itemWithIdentifier:(NSString *)identifier label:(NSString *)label paletteLabel:(NSString *)pallabel segments:(int)segments;

@end

