#import "XeeController.h"
#import "XeeView.h"
#import "XeeImage.h"
#import "XeeImageSaver.h"
#import "XeeSimpleLayout.h"
#import "XeeStatusBar.h"



// XeeController

@implementation XeeController

static float zoomlevels[]={0.03125,0.044,0.0625,0.09,0.125,0.18,0.25,0.35,0.5,0.70,1,1.5,2,3,4,6,8,11,16,23,32};
static int num_zoom=21;

-(id)init
{
	if(self=[super init])
	{
		currimage=nil;
		toolbar=nil;
		toolbaritems=nil;
		toolbaridentifiers=nil;
		defaultidentifiers=nil;

		blocked=NO;
	}
    return self;
}

-(void)dealloc
{
//NSLog(@"XeeController dealloc");

	[currimage setAnimating:NO];
	[currimage release];

	[toolbar release];
	[toolbaritems release];
	[toolbaridentifiers release];
	[defaultidentifiers release];

//	[window autorelease];

	[super dealloc];
}

-(void)awakeFromNib
{
	[self setupStatusBar];
	[self setStatusBarHidden:[[NSUserDefaults standardUserDefaults] boolForKey:[@"hideStatusBar." stringByAppendingString:[[self class] description]]]];

	[self updateWindowPosition];

	[self setupToolbarItems];
	if(toolbaritems)
	{
		toolbar=[[NSToolbar alloc] initWithIdentifier:[[self class] description]];
		[toolbar setDelegate:self];
		[toolbar setDisplayMode:NSToolbarDisplayModeIconOnly];
		[toolbar setAllowsUserCustomization:YES];
		[toolbar setAutosavesConfiguration:YES];
		[window setToolbar:toolbar];
	}
}



-(void)windowWillClose:(NSNotification *)notification
{
//	[window setDelegate:nil];
//	[self release];
}

-(void)windowDidMove:(NSNotification *)notification
{
	[self updateWindowPosition];
}

-(void)windowDidResize:(NSNotification *)notification
{
	[self updateWindowPosition];
}

-(void)windowWillMiniaturize:(NSNotification *)notification
{
	[imageview copyGLtoQuartz];
	[window setOpaque:NO]; // required to make the Quartz underlay and the window shadow appear correctly
}

-(void)windowDidMiniaturize:(NSNotification *)notification
{
	[window setOpaque:YES];
}



-(NSToolbarItem *)toolbar:(NSToolbar *)toolbar itemForItemIdentifier:(NSString *)identifier willBeInsertedIntoToolbar:(BOOL)flag 
{
	return [toolbaritems objectForKey:identifier];
}

-(NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar *)toolbar
{
	return toolbaridentifiers;
}

-(NSArray *)toolbarDefaultItemIdentifiers:(NSToolbar *)toolbar
{
	if(!defaultidentifiers) defaultidentifiers=[[self makeDefaultToolbarItemIdentifiers] retain];
	return defaultidentifiers;
}

-(void)setupToolbarItems
{
	NSArray *items=[self makeToolbarItems];
	if(!items) return;

	NSEnumerator *enumerator;
	NSToolbarItem *item;

	NSMutableDictionary *dict=[NSMutableDictionary dictionaryWithCapacity:[items count]];

	enumerator=[items objectEnumerator];
	while(item=[enumerator nextObject]) [dict setObject:item forKey:[item itemIdentifier]];
	toolbaritems=[[NSDictionary alloc] initWithDictionary:dict];

	NSMutableArray *array=[NSMutableArray arrayWithCapacity:[items count]+3];

	enumerator=[items objectEnumerator];
	while(item=[enumerator nextObject]) [array addObject:[item itemIdentifier]];
	[array addObject:NSToolbarSeparatorItemIdentifier];
	[array addObject:NSToolbarSpaceItemIdentifier];
	[array addObject:NSToolbarFlexibleSpaceItemIdentifier];
	toolbaridentifiers=[[NSArray alloc] initWithArray:array];
}

-(NSArray *)makeToolbarItems { return nil; }

-(NSArray *)makeDefaultToolbarItemIdentifiers { return nil; }



-(BOOL)validateMenuItem:(id <NSMenuItem>)item
{
	return [self validateAction:[item action]];
}

-(BOOL)validateAction:(SEL)action
{
	if(action==@selector(toggleAnimation:)) return currimage&&[currimage animated];
	else if(action==@selector(frameSkipNext:)) return currimage&&[currimage frames]>1;
	else if(action==@selector(frameSkipPrev:)) return currimage&&[currimage frames]>1;
	else if(action==@selector(zoomIn:)) return currimage&&zoom<zoomlevels[num_zoom-1];
	else if(action==@selector(zoomOut:)) return currimage&&zoom>zoomlevels[0];
	else if(action==@selector(zoomActual:)) return currimage&&zoom!=1;
	else if(action==@selector(zoomFit:)) return currimage?YES:NO; //eek, no proper validation
	else return YES;
}



-(void)setupStatusBar { }

-(void)updateStatusBar { }

-(void)setStatusBarHidden:(BOOL)hidden
{
	NSRect imageframe=[imageview frame];
	NSRect statusframe=[statusbar frame];

	if([statusbar isHidden])
	{
		if(hidden) return;
		[statusbar setHidden:NO];
		imageframe.size.height-=statusframe.size.height;
		imageframe.origin.y+=statusframe.size.height;
		[imageview setFrame:imageframe];
		[imageview setDrawResizeCorner:NO];
	}
	else
	{
		if(!hidden) return;
		[statusbar setHidden:YES];
		imageframe.size.height+=statusframe.size.height;
		imageframe.origin.y-=statusframe.size.height;
		[imageview setFrame:imageframe];
		[imageview setDrawResizeCorner:YES];
	}
}

-(BOOL)isStatusBarHidden
{
	return [statusbar isHidden];
}

-(IBAction)toggleStatusBar:(id)sender
{
	BOOL newstate=![statusbar isHidden];
	NSEnumerator *enumerator=[[[NSApplication sharedApplication] windows] objectEnumerator];
	NSWindow *win;

	while(win=[enumerator nextObject])
	{
		id delegate=[win delegate];
		if([delegate class]==[self class]) [delegate setStatusBarHidden:newstate];
	}

	[[NSUserDefaults standardUserDefaults] setBool:newstate forKey:[@"hideStatusBar." stringByAppendingString:[[self class] description]]];
}



-(NSWindow *)window { return window; }



-(void)setImage:(XeeImage *)image
{
	if(image!=currimage)
	{
		[currimage setAnimating:NO];
		[currimage setUpdateAction:nil target:nil];

		[currimage release];
		currimage=[image retain];

		[currimage setUpdateAction:@selector(imageUpdated:) target:self];
		[currimage setAnimatingDefault];
	}

	if(currimage)
	{
		[currimage setFrame:0];
		[imageview setImage:currimage];
		[self setStandardViewSize];
		[imageview setFocus:NSMakePoint(0,0)];
	}
	else
	{
		[imageview setImage:nil];
		//[self setViewSize:NSMakeSize(384,1)];
	}

	[self updateStatusBar];
}

-(void)imageUpdated:(XeeImage *)image
{
	[imageview setNeedsDisplay:YES];
	[self updateStatusBar];
}

-(void)setZoom:(float)newzoom
{
	if(!currimage) return;

	NSSize newsize=NSMakeSize(floor(newzoom*(float)[currimage width]+0.5),floor(newzoom*(float)[currimage height]+0.5));
	[imageview setImageSize:newsize];
	[self setViewSize:newsize];

	zoom=newzoom;

	[self updateStatusBar];
}

-(void)setFrame:(int)frame
{
	if(!currimage) return;
	if(frame==[currimage frame]) return;
	if([currimage frames]==0) return;

	if(frame<0) frame=0;
	if(frame>=[currimage frames]) frame=[currimage frame]-1;

	int oldwidth=[currimage width];
	int oldheight=[currimage height];

	[currimage setFrame:frame];

	if([currimage width]!=oldwidth||[currimage height]!=oldheight)
	{
		NSSize newsize=NSMakeSize(zoom*(float)[currimage width],zoom*(float)[currimage height]);
		[imageview setImageSize:newsize];
		[self setViewSize:newsize];

		[imageview setFocus:NSMakePoint(0,0)];
	}

	[imageview setNeedsDisplay:YES];
	[self updateStatusBar];
}

-(void)updateWindowPosition
{
	NSRect windowframe=[window frame];
	window_focus_x=windowframe.origin.x+windowframe.size.width/2;
	window_focus_y=windowframe.origin.y+windowframe.size.height/2;
}

-(void)setViewSize:(NSSize)size
{
	if([self isResizeBlocked])
	{
		[self setResizeBlock:NO];
		return;
	}

	NSRect screenframe=[self availableScreenSpace];
	NSRect windowframe=[window frame];
	NSSize viewsize=[imageview bounds].size;
	NSSize minsize=[self minViewSize];

	if(size.width<minsize.width) size.width=minsize.width;
	if(size.height<minsize.height) size.height=minsize.height;

	int borderwidth=windowframe.size.width-viewsize.width;
	int borderheight=windowframe.size.height-viewsize.height;
	int win_width=size.width+borderwidth;
	int win_height=size.height+borderheight;

	if(win_width>screenframe.size.width) win_width=screenframe.size.width;
	if(win_height>screenframe.size.height) win_height=screenframe.size.height;

	int focus_x=window_focus_x;
	int focus_y=window_focus_y;
	int win_x=window_focus_x-win_width/2;
	int win_y=window_focus_y-win_height/2;

	if(win_x<screenframe.origin.x) win_x=screenframe.origin.x;
	if(win_y<screenframe.origin.y) win_y=screenframe.origin.y;
	if(win_x+win_width>screenframe.origin.x+screenframe.size.width) win_x=screenframe.origin.x+screenframe.size.width-win_width;
	if(win_y+win_height>screenframe.origin.y+screenframe.size.height) win_y=screenframe.origin.y+screenframe.size.height-win_height;

//	int width=win_width-borderwidth;
//	int height=win_height-borderheight;
//	[imageview setFocus:focus];

	[window setFrame:NSMakeRect(win_x,win_y,win_width,win_height) display:YES];
	[window invalidateCursorRectsForView:imageview]; // just to make sure

	window_focus_x=focus_x; // make sure we remember the old position
	window_focus_y=focus_y;
}

-(void)setStandardViewSize
{
	NSSize maxsize=[self maxViewSize];

	BOOL shrink=[[NSUserDefaults standardUserDefaults] boolForKey:@"shrinkToFit"];
	BOOL enlarge=[[NSUserDefaults standardUserDefaults] boolForKey:@"enlargeToFit"];

	float horiz_zoom=maxsize.width/(float)[currimage width];
	float vert_zoom=maxsize.height/(float)[currimage height];
	float min_zoom=horiz_zoom<vert_zoom?horiz_zoom:vert_zoom;

	zoom=1;
	if(shrink&&min_zoom<1) zoom=min_zoom;
	if(enlarge&&min_zoom>1) zoom=min_zoom;

//	[imageview setImage:currimage];

	NSSize newsize=NSMakeSize(zoom*(float)[currimage width],zoom*(float)[currimage height]);
	[imageview setImageSize:newsize];
	[self setViewSize:newsize];
}

-(void)setResizeBlock:(BOOL)block { blocked=block; }

-(void)setResizeBlockFromSender:(id)sender
{
	if(sender&&[sender isKindOfClass:[NSToolbarItem class]]) [self setResizeBlock:YES];
	else [self setResizeBlock:NO];
}

-(BOOL)isResizeBlocked
{
	switch([[NSUserDefaults standardUserDefaults] integerForKey:@"windowResizing"])
	{
		case 1: return blocked;
		case 2: return YES;
	}
	return NO;
}

-(NSSize)maxViewSize
{
	if([self isResizeBlocked])
	{
		return [imageview bounds].size;
	}
	else
	{
		NSSize screensize=[self availableScreenSpace].size;
		NSSize windowsize=[window frame].size;
		NSSize viewsize=[imageview bounds].size;

		return NSMakeSize(screensize.width-windowsize.width+viewsize.width,screensize.height-windowsize.height+viewsize.height);
	}
}

-(NSSize)minViewSize
{
	return NSMakeSize(256,128);
}

-(NSRect)availableScreenSpace
{
	return [[window screen] visibleFrame];
}



-(IBAction)copy:(id)sender
{
	if(currimage)
	{
		NSData *tiff=[[currimage imageRep] TIFFRepresentation];
		NSPasteboard *pboard=[NSPasteboard generalPasteboard];

		[pboard declareTypes:[NSArray arrayWithObject:NSTIFFPboardType] owner:nil];
		[pboard setData:tiff forType:NSTIFFPboardType];
	}
}

-(IBAction)saveImage:(id)sender
{
	if(currimage) [XeeSavePanel runSavePanelForImage:currimage window:window];
}



-(IBAction)frameSkipNext:(id)sender
{
	[self setResizeBlockFromSender:sender];
	if(currimage)
	{
		int frame=[currimage frame];
		int frames=[currimage frames];
		[self setFrame:(frame+1)%frames];
	}
}

-(IBAction)frameSkipPrev:(id)sender
{
	[self setResizeBlockFromSender:sender];
	if(currimage)
	{
		int frame=[currimage frame];
		int frames=[currimage frames];
		[self setFrame:(frame+frames-1)%frames];
	}
}

-(IBAction)toggleAnimation:(id)sender
{
	if(!currimage||![currimage animated]) return;

	[currimage setAnimating:![currimage animating]];
}



-(IBAction)zoomIn:(id)sender
{
	[self setResizeBlockFromSender:sender];

	int i;
	for(i=0;i<num_zoom-1;i++) if(zoomlevels[i]>zoom) break;

	[self setZoom:zoomlevels[i]];
}

-(IBAction)zoomOut:(id)sender
{
	[self setResizeBlockFromSender:sender];

	int i;
	for(i=num_zoom-1;i>0;i--) if(zoomlevels[i]<zoom) break;

	[self setZoom:zoomlevels[i]];
}

-(IBAction)zoomActual:(id)sender
{
	[self setResizeBlockFromSender:sender];

	[self setZoom:1];
}

-(IBAction)zoomFit:(id)sender
{
	[self setResizeBlockFromSender:sender];

	NSSize maxsize=[self maxViewSize];

	float horiz_zoom=maxsize.width/(float)[currimage width];
	float vert_zoom=maxsize.height/(float)[currimage height];

	[self setZoom:horiz_zoom<vert_zoom?horiz_zoom:vert_zoom];
}

-(IBAction)setAutoZoom:(id)sender
{
	[self setResizeBlockFromSender:sender];
	[self setStandardViewSize];
}



-(IBAction)fullScreen:(id)sender
{
/*	[NSMenu setMenuBarVisible:NO];
	[self setStatusBarHidden:YES];
	[window setFrame:[window frameRectForContentRect:[[window screen] frame]] display:YES];
//	[window set NSBorderlessWindowMask:
*/

	//[[imageview openGLContext] setFullScreen:YES];

}

-(IBAction)closeWindowOrDrawer:(id)sender
{
//	[[NSApplication sharedApplication] sendAction:@selector(performClose:) to:nil from:self];
	[[self window] performClose:self];
}

@end




@implementation XeeWindow

-(void)dealloc
{
//NSLog(@"XeeWindow dealloc");
	[[self delegate] release];
	[super dealloc];
}

@end


/*
@implementation NSToolbarItem (QuickConstructor)

+(NSToolbarItem *)itemWithIdentifier:(NSString *)ident label:(NSString *)label image:(NSString *)imagename action:(SEL)action
{
	NSToolbarItem *item=[[NSToolbarItem alloc] initWithItemIdentifier:ident];
	[item setLabel:label];
	[item setImage:[NSImage imageNamed:imagename]];
	[item setAction:action];
	return [item autorelease];
}

+(NSToolbarItem *)itemWithIdentifier:(NSString *)ident label:(NSString *)label view:(NSView *)view
{
	NSToolbarItem *item=[[NSToolbarItem alloc] initWithItemIdentifier:ident];
	[item setLabel:label];
	//[view removeFromSuperview];
	[item setView:view];
	[item setMinSize:[view bounds].size];
	[item setMaxSize:[view bounds].size];
	return [item autorelease];
}

@end
*/


@implementation XeeSegmentedItem

-(id)initWithItemIdentifier:(NSString *)identifier label:(NSString *)label paletteLabel:(NSString *)pallabel segments:(int)segments
{
	if(self=[super initWithItemIdentifier:identifier])
	{
		[self setLabel:label];
		[self setPaletteLabel:pallabel];

		control=[[NSSegmentedControl alloc] init];
		[control setSegmentCount:segments];
		[[control cell] setTrackingMode:NSSegmentSwitchTrackingMomentary];

		if(segments!=1)
		{
			menu=[[NSMenu alloc] init];

			for(int i=0;i<segments;i++)
			[menu addItem:[[[NSMenuItem alloc] initWithTitle:@"" action:NULL keyEquivalent:@""] autorelease]];

			NSMenuItem *item=[[[NSMenuItem alloc] initWithTitle:pallabel action:NULL keyEquivalent:@""] autorelease];
			[item setSubmenu:menu];

			[self setMenuFormRepresentation:item];
		}
		else
		{
			NSMenuItem *item=[[[NSMenuItem alloc] initWithTitle:@"" action:NULL keyEquivalent:@""] autorelease];
			[self setMenuFormRepresentation:item];
			menu=nil;
		}

		[control setTarget:self];
		[control setAction:@selector(clicked:)];

		actions=malloc(sizeof(SEL)*segments);
	}
	return self;
}

-(void)dealloc
{
	[control release];
	[menu release];

	free(actions);

	[super dealloc];
}

-(void)validate
{
	if([[NSApplication sharedApplication] keyWindow]!=[control window]) [self setEnabled:NO];
	else
	{
		[self setEnabled:YES];

		int count=[control segmentCount];
		for(int i=0;i<count;i++)
		{
			XeeController *controller=(XeeController *)[[control window] delegate];
			[control setEnabled:[controller validateAction:actions[i]] forSegment:i];
		}
	}
}

-(void)setSegment:(int)segment label:(NSString *)label image:(NSImage *)image longLabel:(NSString *)longlabel width:(int)width action:(SEL)action
{
	if(segment<0 || segment>=[control segmentCount]) return;

	[control setLabel:label forSegment:segment];
	[control setImage:image forSegment:segment];
	[control setWidth:width forSegment:segment];
	[[control cell] setToolTip:longlabel forSegment:segment];

	actions[segment]=action;

	NSMenuItem *item;
	if(menu) item=[menu itemAtIndex:segment];
	else item=[self menuFormRepresentation];

	[item setTitle:longlabel];
	[item setImage:image];
	[item setAction:action];
}

-(void)setSegment:(int)segment label:(NSString *)label longLabel:(NSString *)longlabel action:(SEL)action
{
	[self setSegment:segment label:label image:nil longLabel:longlabel width:0 action:action];
}

-(void)setSegment:(int)segment imageName:(NSString *)imagename longLabel:(NSString *)longlabel action:(SEL)action
{
	NSImage *image=[NSImage imageNamed:imagename];
	[self setSegment:segment label:nil image:image longLabel:longlabel width:[image size].width action:action];
}

-(void)setupView
{
	[control sizeToFit];
	[self setView:control];
	[self setMinSize:[control frame].size];
	[self setMaxSize:[control frame].size];
}

-(void)clicked:(id)sender
{
	[[NSApplication sharedApplication] sendAction:actions[[sender selectedSegment]] to:nil from:self];
}

+(XeeSegmentedItem *)itemWithIdentifier:(NSString *)identifier label:(NSString *)label paletteLabel:(NSString *)pallabel segments:(int)segments
{
	return [[[XeeSegmentedItem alloc] initWithItemIdentifier:identifier label:label paletteLabel:pallabel segments:segments] autorelease];
}

@end


@implementation NSSegmentedCell (AlwaysTextured)

-(BOOL)_isTextured  { return YES; }

@end
