#import "XeeILBMLoader.h"
#import "XeeIFFReader.h"



#define PARSE_ID(ptr) (((ptr)[0]<<24)|((ptr)[1]<<16)|((ptr)[2]<<8)|(ptr)[3])



@implementation XeeILBMImage

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes
{
	if(self=[super initWithFile:name firstBlock:block attributes:attributes])
	{
		image=NULL;
		mask=NULL;
		iff=nil;
		ranges=[[NSMutableArray alloc] initWithCapacity:4];
		clock=0;
		animationtimer=nil;

		byte *header=(byte *)[block bytes];

		if([block length]>12&&PARSE_ID(header)=='FORM'&&PARSE_ID(header+8)=='ILBM')
		{
			iff=[[XeeIFFReader alloc] initWithFilename:name filetype:'ILBM'];
			if(iff)
			{
				int num,viewmode,col;
				XeeILBMRange *range;
				BOOL reading=YES;

				viewmode=0;
				ham=ham8=ehb=ocscol=NO;

				while(reading)
				switch([iff nextChunk])
				{
					case 'BMHD':
						realwidth=[iff readWord];
						realheight=[iff readWord];
						[iff readWord]; // x
						[iff readWord]; // y
						planes=[iff readByte];
						masking=[iff readByte];
						compression=[iff readByte];
						[iff readByte]; // pad1
						trans=[iff readWord];
						xasp=[iff readByte];
						yasp=[iff readByte];
						// pagewidth, pageheight;

						if(xasp>yasp) { xscale=(float)xasp/(float)yasp+0.5; yscale=1; }
						else { yscale=(float)yasp/(float)xasp+0.5; xscale=1; }

						if(xscale>4) xscale=4;
						if(yscale>4) yscale=4;

						if(planes<=8)
						{
							width=realwidth*xscale;
							height=realheight*yscale;
						}
						else // no scaling support for 24 bit
						{
							width=realwidth;
							height=realheight;
						}

						rowbytes=2*((realwidth+15)/16);

						if(planes>8&&planes!=24) reading=NO;

						//printf("w:%d h:%d planes:%d mask:%d comp:%d transp:%d xasp:%d yasp:%d xscale:%d yscale:%d\n",width,height,planes,masking,compression,trans,xasp,yasp,xscale,yscale);
					break;

					case 'CMAP':
						num=[iff chunkSize]/3;
						if(num>256) num=256;

						col=0;

						for(int i=0;i<num;i++)
						{
							byte r=[iff readByte];
							byte g=[iff readByte];
							byte b=[iff readByte];
							palette[i]=0xff000000|(r<<16)|(g<<8)|b;
							if(i<32) col|=palette[i];
						}

						if((col&0x0f0f0f)==0) ocscol=YES;
					break;

					case 'DEST':
						printf("warning: image has DEST chunk\n");
					break;

					case 'CRNG':
					case 'DRNG':
					case 'CCRT':
						range=[[[XeeILBMRange alloc] initWithIFF:iff image:self] autorelease];
						if(range) [ranges addObject:range];
					break;

					case 'CAMG':
						viewmode=[iff readLong];
					break;

					case 'BODY':
						if((viewmode&0x800)&&(planes==5||planes==6)) ham=YES;
						else if((viewmode&0x800)&&(planes==7||planes==8)) ham8=YES;
						else if(viewmode&0x80) ehb=YES;

						if(ham)
						{
							[self setDepth:@"Amiga HAM"];
							[self setDepthIconName:@"depth_amiga"];
						}	
						else if(ham8)
						{
							[self setDepth:@"Amiga HAM8"];
							[self setDepthIconName:@"depth_amiga"];
						}
						else if(ehb)
						{
							[self setDepth:@"Amiga EHB"];
							[self setDepthIconName:@"depth_amiga"];
						}
						else if(planes==24) [self setDepthRGB:24];
						else [self setDepthIndexed:1<<planes];

						if(ocscol&&(planes<=5||ehb||ham)) // fix 4-bit OCS palette
						{
							for(int i=0;i<32;i++) palette[i]|=(palette[i]>>4)&0x0f0f0f;
						}

						if(ehb) // fix EHB palette
						{
							for(int i=0;i<32;i++) palette[i+32]=((palette[i]>>1)&0x7f7f7f)|0xff000000;
						}

						if(masking)
						{
							int col=palette[trans];
	
							[self setBackgroundColor:[NSColor
							colorWithCalibratedRed:(float)((col>>16)&0xff)/255.0
							green:(float)((col>>8)&0xff)/255.0
							blue:(float)(col&0xff)/255.0
							alpha:1]];

							if(masking==1&&[[NSUserDefaults standardUserDefaults] boolForKey:@"ilbmUseMask"])
							{
								[self setTransparent:YES];
							}
							else if(masking==2&&[[NSUserDefaults standardUserDefaults] boolForKey:@"ilbmUseTransparentColour"])
							{
								palette[trans]&=0xffffff;
								[self setTransparent:YES];
							}
						}

						[ranges makeObjectsPerformSelector:@selector(setup) withObject:nil];

						[self setFormat:@"IFF-ILBM"];

						[self setState:XeeImageStateInit];
						return self;
					break;

					case 0:
						reading=NO;
					break;
				}
			}
		}

		[self release];
	}

	return nil;
}

-(void)dealloc
{
	[iff release];
	[ranges release];

	if(image) free(image);
	if(mask) free(mask);

	[animationtimer release];

	[super dealloc];
}

-(void)load
{
	if([self state]==XeeImageStateInit)
	{
		if(planes<=8)
		{
			image=malloc(realwidth*realheight);
			if(!image)
			{
				[self setState:XeeImageStateFailed];
				return;
			}
			memset(image,0,realwidth*realheight);
		}

		if(masking==1)
		{
			mask=malloc(rowbytes*realheight);
			if(!mask)
			{
				[self setState:XeeImageStateFailed];
				return;
			}
		}

		if(![self allocBitmapWithType:XeeBitmapTypeARGB32 width:width height:height])
		{
			[self setState:XeeImageStateFailed];
			return;
		}

		current_line=0;

		[self setState:XeeImageStateLoading];
	}

	stop=NO;

	while(!stop)
	{
		if(planes<=8)
		{
			byte *imageline=image+current_line*realwidth;
			byte row[rowbytes];

			for(int p=0;p<planes;p++)
			{
				if(![self readRow:row])
				{
					[self setState:XeeImageStateFailed];
					goto end;
				}

				for(int x=0;x<realwidth;x++)
				{
					imageline[x]|=((row[x>>3]>>((x&7)^7))&1)<<p;
				}
			}

			if(masking==1)
			{
				if(![self readRow:mask+current_line*rowbytes])
				{
					[self setState:XeeImageStateFailed];
					goto end;
				}
			}
		}
		else if(planes==24)
		{
			int *imageline=(int *)(data+current_line*bytesperrow);
			byte row[rowbytes];

			for(int i=0;i<width;i++) imageline[i]=0xff000000;

			for(int p=0;p<planes;p++)
			{
				if(![self readRow:row])
				{
					[self setState:XeeImageStateFailed];
					goto end;
				}

				for(int i=0;i<width;i++)
				{
					imageline[i]|=((row[i>>3]>>((i&7)^7))&1)<<(23-(p^7));
				}
			}

			if(masking==1)
			{
				if(![self readRow:row])
				{
					[self setState:XeeImageStateFailed];
					goto end;
				}

				for(int i=0;i<width;i++)
				{
					if(!((row[i>>3]>>((i&7)^7))&1)) imageline[i]=0;
				}
			}
		}

		current_line++;
		if(current_line>=realheight) break;
	}
	end:

	[self renderImage];

	if(stop) return;

	[iff release];
	iff=nil;

	if(current_line>=realheight) [self setState:XeeImageStateLoaded];

	[self triggerUpdateAction];
}

-(BOOL)readRow:(byte *)row
{
	if(compression==0)
	{
		for(int i=0;i<rowbytes;i++)
		{
			if(![iff bytesLeft]) return NO;
			row[i]=[iff readByte];
		}
	}
	else
	{
		int count=0;
		while(count<rowbytes)
		{
			if(![iff bytesLeft]) return NO;
			int num=[iff readByte];

			if(num<128)
			{
				num+=1;

				for(int i=0;i<num;i++)
				{
					if(count<rowbytes)
					{
						if(![iff bytesLeft]) return NO;
						row[count]=[iff readByte];
					}
					count++;
				}
			}
			else if(num>128)
			{
				if(![iff bytesLeft]) return NO;
				byte b=[iff readByte];

				num=257-num;

				if(count+num>rowbytes) num=rowbytes-count;
				for(int i=0;i<num;i++) row[count++]=b;
			}
		}
	}

	return YES;
}

-(void)renderImage
{
	if(planes==24) return;

	if(ham)
	{
		for(int y=0;y<realheight;y++)
		for(int sub_y=0;sub_y<yscale;sub_y++)
		{
			byte *srcline=image+y*realwidth;
			int *destline=(int *)(data+(y*yscale+sub_y)*bytesperrow);
			unsigned long hold=palette[0];

			for(int x=0;x<realwidth;x++)
			{
				int pixel=*srcline++;
				int cmd=pixel>>4;
				int val=pixel&0x0f;

				switch(cmd)
				{
					case 0: hold=palette[val]; break;
					case 1: hold=(hold&0xffffff00)|(val<<4)|(val); break;
					case 2: hold=(hold&0xff00ffff)|(val<<20)|(val<<16); break;
					case 3: hold=(hold&0xffff00ff)|(val<<12)|(val<<8); break;
				}

				for(int sub_x=0;sub_x<xscale;sub_x++) *destline++=hold;
			}
		}
	}
	else if(ham8)
	{
		for(int y=0;y<realheight;y++)
		for(int sub_y=0;sub_y<yscale;sub_y++)
		{
			byte *srcline=image+y*realwidth;
			int *destline=(int *)(data+(y*yscale+sub_y)*bytesperrow);
			unsigned long hold=palette[0];

			for(int x=0;x<realwidth;x++)
			{
				int pixel=*srcline++;
				int cmd=pixel>>6;
				int val=pixel&0x3f;

				switch(cmd)
				{
					case 0: hold=palette[val]; break;
					case 1: hold=(hold&0xffffff00)|(val<<2)|((val&0x30)>>4); break;
					case 2: hold=(hold&0xff00ffff)|(val<<18)|((val&0x30)<<12); break;
					case 3: hold=(hold&0xffff00ff)|(val<<10)|((val&0x30)<<4); break;
				}

				for(int sub_x=0;sub_x<xscale;sub_x++) *destline++=hold;
			}
		}
	}
	else
	{
		for(int y=0;y<realheight;y++)
		for(int sub_y=0;sub_y<yscale;sub_y++)
		{
			byte *srcline=image+y*realwidth;
			int *destline=(int *)(data+(y*yscale+sub_y)*bytesperrow);

			for(int x=0;x<realwidth;x++)
			{
				unsigned long col=palette[*srcline++];
				for(int sub_x=0;sub_x<xscale;sub_x++) *destline++=col;
			}
		}
	}

	if(masking==1&&[[NSUserDefaults standardUserDefaults] boolForKey:@"ilbmUseMask"])
	{
		for(int y=0;y<realheight;y++)
		for(int sub_y=0;sub_y<yscale;sub_y++)
		{
			byte *maskline=mask+y*rowbytes;
			int *destline=(int *)(data+(y*yscale+sub_y)*bytesperrow);

			for(int x=0;x<realwidth;x++)
			{
				if(!((maskline[x>>3]>>((x&7)^7))&1))
				{
					for(int sub_x=0;sub_x<xscale;sub_x++) destline[sub_x]=0;
				}
				destline+=xscale;
			}
		}
	}
}

-(unsigned long *)palette { return palette; }




-(BOOL)animated { return [ranges count]?YES:NO; }

-(void)setAnimating:(BOOL)animating
{
	if(![self animated]) return;

	if(animating)
	{
		if(!animationtimer)
		{
			animationtimer=[[NSTimer scheduledTimerWithTimeInterval:1.0/60.0 target:self selector:@selector(animate:) userInfo:nil repeats:YES] retain];
		}
	}
	else
	{
		if(animationtimer)
		{
			[animationtimer invalidate];
			[animationtimer release];
			animationtimer=nil;
		}
	}
}

-(BOOL)animating { return animationtimer?YES:NO; }

-(void)animate:(NSTimer *)timer
{
	clock++;

	NSEnumerator *enumerator=[ranges objectEnumerator];
	XeeILBMRange *range;
	BOOL triggered=NO;

	while(range=[enumerator nextObject])
	{
		if([range triggerCheck:(float)clock/60.0]) triggered=YES;
	}

	if(triggered)
	{
		[self renderImage];
		[self triggerUpdateAction];
	}
}



+(NSArray *)fileTypes
{
	return [NSArray arrayWithObjects:@"iff",@"ilbm",@"lbm",@"'ILBM'",nil]; //@"'PNG '",@"'PNGf'",nil];
}

@end



@implementation XeeILBMRange

-(id)initWithIFF:(XeeIFFReader *)iff image:(XeeILBMImage *)img;
{
	if(self=[super init])
	{
		num=0;
		colours=NULL;
		indexes=NULL;
		image=img;

		switch([iff currentChunk])
		{
			case 'CRNG':
			{
				[iff readWord]; // pad1
				int rate=[iff readWord];
				int flags=[iff readWord];
				int low=[iff readByte];
				int high=[iff readByte];

				if((flags&1)&&rate&&low<high)
				{
					if([self allocBuffers:high-low+1])
					{
						next=interval=16384.0/(float)rate/60.0;

						[self setIndexesFrom:low to:high reverse:(flags&2)?YES:NO];

						return self;
					}
				}
			}
			break;

			case 'CCRT':
			{
				int dir=[iff readWord];
				int start=[iff readByte];
				int end=[iff readByte];
				int secs=[iff readLong];
				int micros=[iff readLong];

				if((dir==1||dir==2)&&start<end)
				{
					if([self allocBuffers:end-start+1])
					{
						next=interval=(float)secs+(float)micros/1000000.0;

						[self setIndexesFrom:start to:end reverse:dir==2]; // unsure if this should be ==2 or ==1.

						return self;
					}
				}
			}
			break;

			case 'DRNG':
			{
				int min=[iff readByte];
				int max=[iff readByte];
				int rate=[iff readWord]*100;
				int flags=[iff readWord];
				int ntrue=[iff readByte];
				int nregs=[iff readByte];

				//printf("DRNG min:%d max:%d rate:%x flags:%d ntrue:%d nregs:%d\n",min,max,rate,flags,ntrue,nregs);

				if((flags&1)&&rate&&min<max)
				{
					if([self allocBuffers:max-min+1])
					{
						for(int i=0;i<ntrue;i++)
						{
							byte cell=[iff readByte];
							byte r=[iff readByte];
							byte g=[iff readByte];
							byte b=[iff readByte];

							if(cell>=min&&cell<=max) colours[cell-min]=0xff000000|(r<<16)|(g<<8)|b;
						}

						for(int i=0;i<nregs;i++)
						{
							byte cell=[iff readByte];
							byte index=[iff readByte];

							if(cell>=min&&cell<=max) indexes[cell-min]=index;
						}

						next=interval=16384.0/(float)rate/60.0;

						return self;
					}
				}
			}
			break;
		}

		[self release];
	}
	return nil;
}

-(void)dealloc
{
	if(colours) free(colours);
	if(indexes) free(indexes);

	[super dealloc];
}

-(BOOL)allocBuffers:(int)length
{
	if(colours) free(colours);
	if(indexes) free(indexes);

	colours=malloc(sizeof(unsigned long)*length);
	indexes=malloc(sizeof(int)*length);

	if(colours&&indexes)
	{
		num=length;

		for(int i=0;i<num;i++)
		{
			colours[i]=0x80000000;
			indexes[i]=-1;
		}

		return YES;
	}
	else
	{
		num=0;
		return NO;
	}
}

-(void)setIndexesFrom:(int)start to:(int)end reverse:(BOOL)reverse
{
	if(reverse)
	for(int i=0;i<=end-start;i++) indexes[i]=i+start;
	else
	for(int i=0;i<=end-start;i++) indexes[i]=end-i;
}

-(void)setup
{
	unsigned long *palette=[image palette];

	for(int i=0;i<num;i++)
	{
		if(colours[i]==0x80000000&&indexes[i]>=0) colours[i]=palette[indexes[i]];
	}

	for(int i=0;i<num;i++)
	{
		if(colours[i]==0x80000000)
		{
			unsigned long startindex,endindex;
			int startcol,endcol;

			if(i!=0) startcol=colours[i-1];

			startindex=i-1;
			endindex=i+1;

			while(endindex<num&&colours[endindex]==0x80000000) endindex++;

			if(endindex<num)
			{
				endcol=colours[endindex];
				if(startindex>=0) startcol=colours[startindex];
				else startcol=endcol;
			}
			else 
			{
				if(startindex>=0) endcol=startcol=colours[startindex];
				else startcol=endcol=0xff000000;
			}

			int start_r=(startcol>>16)&0xff;
			int start_g=(startcol>>8)&0xff;
			int start_b=startcol&0xff;
			int end_r=(endcol>>16)&0xff;
			int end_g=(endcol>>8)&0xff;
			int end_b=endcol&0xff;
			int total=endindex-startindex;

			for(int j=startindex+1;j<endindex;j++)
			{
				int step=j-startindex;
				int r=start_r+((end_r-start_r)*step)/total;
				int g=start_g+((end_g-start_g)*step)/total;
				int b=start_b+((end_b-start_b)*step)/total;

				colours[j]=0xff000000|(r<<16)|(g<<8)|b;
			}

			i=endindex;
		}
	}
}

-(BOOL)triggerCheck:(float)time
{
	if(time>next)
	{
		while(time>next)
		{
			[self cycle];
			next+=interval;
		}
		return YES;
	}
	return NO;
}


-(void)cycle
{
	unsigned long *palette=[image palette];
	unsigned long tmp=colours[0];

	for(int i=1;i<num;i++) colours[i-1]=colours[i];

	colours[num-1]=tmp;

	for(int i=0;i<num;i++)
	{
		if(indexes[i]>=0) palette[indexes[i]]=colours[i];
	}
}

@end
