#import <Cocoa/Cocoa.h>
#import "XeeSimpleLayout.h"

@class XeeImage;

@interface XeeSavePanel:NSSavePanel
{
	XeeImage *image;
	NSArray *savers;
	XeeSLPages *formats;
	XeeSimpleLayout *view;

	NSTextView *textview;

	BOOL wasanimating;
}

+(void)runSavePanelForImage:(XeeImage *)image window:(NSWindow *)window;

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;

-(void)beginSheetForWindow:(NSWindow *)window;

-(void)savePanelDidEnd:(NSSavePanel *)sheet returnCode:(int)res contextInfo:(void *)info;
-(void)xeeSLUpdated:(XeeSimpleLayout *)alsoview;

-(NSString *)updateExtension:(NSString *)filename;
-(void)selectNamePart;

@end




@interface XeeImageSaver:NSObject
{
	XeeImage *image;
	XeeSLControl *control;
}

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;

-(NSString *)format;
-(NSString *)extension;
-(BOOL)save:(NSString *)filename;

-(XeeSLControl *)control;
-(void)setControl:(XeeSLControl *)cont;

+(void)initialize;
+(NSArray *)saversForImage:(XeeImage *)image;
+(void)registerSaverClass:(Class)class;

@end



@interface XeeNSImageSaver:XeeImageSaver
{
	NSBitmapImageFileType type;
}

-(id)initWithImage:(XeeImage *)img;
-(BOOL)save:(NSString *)filename;

-(void)setFileType:(NSBitmapImageFileType)filetype;

@end



@interface XeeAlphaSaver:XeeNSImageSaver
{
	XeeSLSwitch *alpha;
}

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;
-(BOOL)save:(NSString *)filename;

@end


@interface XeePNGSaver:XeeAlphaSaver
{
	//XeeSLPopUp *depth;
	XeeSLSwitch *interlaced;
}

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;
-(NSString *)format;
-(NSString *)extension;
-(BOOL)save:(NSString *)filename;

@end



@interface XeeJP2Saver:XeeAlphaSaver
{
	XeeSLSlider *quality;
}

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;
-(NSString *)format;
-(NSString *)extension;
-(BOOL)save:(NSString *)filename;

@end



@interface XeeTIFFSaver:XeeAlphaSaver
{
	//XeeSLPopUp *depth;
	XeeSLPopUp *compression;
}

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;
-(NSString *)format;
-(NSString *)extension;
-(BOOL)save:(NSString *)filename;

@end



@interface XeeBMPSaver:XeeAlphaSaver
{
}

-(id)initWithImage:(XeeImage *)img;
-(NSString *)format;
-(NSString *)extension;

@end



@interface XeeJPEGSaver:XeeNSImageSaver
{
	XeeSLSlider *quality;
}

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;
-(NSString *)format;
-(NSString *)extension;

@end



@interface XeeGIFSaver:XeeNSImageSaver
{
	XeeSLPopUp *transparency;
}

-(id)initWithImage:(XeeImage *)img;
-(void)dealloc;
-(NSString *)format;
-(NSString *)extension;

@end
