#import "XeePNGLoader.h"



@implementation XeePNGImage

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes
{
	if(self=[super initWithFile:name firstBlock:block attributes:attributes])
	{
		png=NULL;
		info=NULL;
		fh=NULL;

		if([block length]>8&&png_check_sig((unsigned char *)[block bytes],8))
		{
			fh=fopen([filename fileSystemRepresentation],"rb");

			if(fh)
			{
				png=png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);

				if(png)
				{
					info=png_create_info_struct(png);

					if(info)
					{
						if(setjmp(png_jmpbuf(png)))
						{
							[self release];
							return nil;
						}

						png_init_io(png,fh);
						//png_set_sig_bytes(png_ptr, 8);  // we already read the 8 signature bytes

						png_read_info(png,info); // read all PNG info up to image data

						width=png_get_image_width(png,info);
						height=png_get_image_height(png,info);
						bit_depth=png_get_bit_depth(png,info);
						color_type=png_get_color_type(png,info);

						switch(color_type)
						{
							case PNG_COLOR_TYPE_GRAY: [self setDepthGrey:(1<<bit_depth)]; break;
							case PNG_COLOR_TYPE_GRAY_ALPHA: [self setDepthGreyAlpha:bit_depth]; break;
							case PNG_COLOR_TYPE_PALETTE: [self setDepthIndexed:(1<<bit_depth)]; break;
							case PNG_COLOR_TYPE_RGB: [self setDepthRGB:3*bit_depth]; break;
							case PNG_COLOR_TYPE_RGB_ALPHA: [self setDepthRGBA:4*bit_depth]; break;
						}

						if(color_type==PNG_COLOR_TYPE_RGB_ALPHA||color_type==PNG_COLOR_TYPE_GRAY_ALPHA||
						png_get_valid(png,info,PNG_INFO_tRNS))
						{
							[self setTransparent:YES];

							if(png_get_valid(png,info,PNG_INFO_bKGD))
							{
								png_color_16p background;
								float scaling=(1<<bit_depth)-1;

								png_get_bKGD(png,info,&background);

								[self setBackgroundColor:[NSColor colorWithCalibratedRed:(float)background->red/scaling
								green:(float)background->green/scaling blue:(float)background->blue/scaling alpha:1]];
							}
							else [self setBackgroundColor:[NSColor whiteColor]];
						}

						[self setFormat:@"PNG"];

						current_line=0;
						current_pass=0;

						[self setState:XeeImageStateInit];

						return self;
					}
				}
			}
		}

		[self release];
	}

	return nil;
}

-(void)dealloc
{
	if(info) png_destroy_read_struct(&png,&info,NULL);
	else if(png) png_destroy_read_struct(&png,NULL,NULL);
	if(fh) fclose(fh);

	[super dealloc];
}

int is_png_gray_palette(png_structp png,png_infop info)
{
	if(png_get_valid(png,info,PNG_INFO_tRNS)) return 0;
	if(png_get_color_type(png,info)!=PNG_COLOR_TYPE_PALETTE) return 0;

	png_colorp pal;
	int num;

	png_get_PLTE(png,info,&pal,&num);
	for(int i=0;i<num;i++) if(pal[i].red!=pal[i].green||pal[i].red!=pal[i].blue) return 0;

	return 1;
}

-(void)load
{
    if(setjmp(png_jmpbuf(png)))
	{
		[self setState:XeeImageStateFailed];
		[self triggerUpdateAction];
		return;
	}

	if([self state]==XeeImageStateInit)
	{
		int bitmaptype;

		switch(color_type)
		{
			case PNG_COLOR_TYPE_GRAY:
				bitmaptype=XeeBitmapTypeLuma8;
				if(bit_depth<8) png_set_expand(png); // expand low-bit-depth grayscale images to 8 bits
			break;

			case PNG_COLOR_TYPE_GRAY_ALPHA:
				bitmaptype=XeeBitmapTypeARGB32;
				png_set_gray_to_rgb(png); // convert GA to RGBA
			break;

			case PNG_COLOR_TYPE_PALETTE:
				if(is_png_gray_palette(png,info))
				{
					bitmaptype=XeeBitmapTypeLuma8;
					png_set_rgb_to_gray_fixed(png,1,-1,-1); // triggers a bug in libpng 1.2.8, needs patched libpng
				}
				else
				{
					bitmaptype=XeeBitmapTypeRGB24;
					png_set_expand(png);  // expand palette images to RGB
				}
			break;

			case PNG_COLOR_TYPE_RGB:
				bitmaptype=XeeBitmapTypeRGB24;
			break;

			case PNG_COLOR_TYPE_RGB_ALPHA:
				bitmaptype=XeeBitmapTypeARGB32;
			break;
		}

		if(bit_depth==16) png_set_strip_16(png);

		if(png_get_valid(png,info,PNG_INFO_tRNS))
		{
			bitmaptype=XeeBitmapTypeARGB32;
			png_set_expand(png); // expand transparency to RGBA
			if(color_type==PNG_COLOR_TYPE_GRAY) png_set_gray_to_rgb(png); // convert G+A to RGBA
		}

		png_set_swap_alpha(png); // make ARGB values instead of RGBA

//		if(png_get_gAMA(png_ptr,info_ptr,&gamma)) png_set_gamma(png_ptr,display_exponent,gamma);

		interlace_passes=png_set_interlace_handling(png);

		png_read_update_info(png,info);

		if(![self allocBitmapWithType:bitmaptype width:width height:height])
		{
			[self setState:XeeImageStateFailed];
			return;
		}

		if(bytesperrow<png_get_rowbytes(png,info)) // shouldn't happen
		{
			[self setState:XeeImageStateFailed];
			return;
		}

		[self setState:XeeImageStateLoading];
	}

	stop=NO;

	while(!stop)
	{
		unsigned char *row=data;
		row+=current_line*bytesperrow;

		png_read_row(png,row,NULL);

		current_line++;
		if(current_line>=height)
		{
			current_line=0;
			current_pass++;
			if(current_pass>=interlace_passes) break;
		}
	}

	if(stop) return;

	png_destroy_read_struct(&png,&info,NULL);
	png=NULL;
	info=NULL;

	fclose(fh);
	fh=NULL;

	[self setState:XeeImageStateLoaded];
	[self triggerUpdateAction];
}

+(NSArray *)fileTypes
{
	return [NSArray arrayWithObjects:@"png",@"'PNG '",@"'PNGf'",nil];
}

@end
