#import "XeeView.h"
#import "XeeImage.h"
#import "XeeController.h"

#import <sys/time.h>


GLuint make_grid_texture(float r,float b,float g);
GLuint make_resize_texture();


@implementation XeeView

-(id)initWithFrame:(NSRect)frameRect
{
/*	NSOpenGLPixelFormatAttribute attrs[]={ 
		NSOpenGLPFANoRecovery,
		NSOpenGLPFADoubleBuffer,
		//NSOpenGLPFAFullScreen,
		NSOpenGLPFAAccelerated,
		(NSOpenGLPixelFormatAttribute)0};

	NSOpenGLPixelFormat *format=[[[NSOpenGLPixelFormat alloc] initWithAttributes:attrs] autorelease];*/

	if(self=[super initWithFrame:frameRect/* pixelFormat:format*/])
	{
		image=nil;
		scrolltimer=nil;

		drawresize=NO;

		up=down=left=right=NO;

		NSRect bounds=[self bounds];
		width=bounds.size.width;
		height=bounds.size.height;

		long val=1;
		[[self openGLContext] setValues:&val forParameter:NSOpenGLCPSwapInterval];

		[self registerForDraggedTypes:[NSArray arrayWithObject:NSFilenamesPboardType]];
	}
	return self;
}

-(void)dealloc
{
	[image setAnimating:NO];
	[image release];
	[scrolltimer release];

	[super dealloc];
}

-(void)drawRect:(NSRect)rect
{
	[[self openGLContext] makeCurrentContext];

	glViewport(0,0,width,height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0,width,height,0);
	glMatrixMode(GL_MODELVIEW);

	glClearColor(0,0,0,1);
	glClear(GL_COLOR_BUFFER_BIT);

	if(image)
	{
		int draw_x,draw_y;

		if(imgwidth<width) draw_x=(width-imgwidth)/2;
		else draw_x=-x;

		if(imgheight<height) draw_y=(height-imgheight)/2;
		else draw_y=-y;

		if([image transparent])
		{
			NSColor *back=[[image backgroundColor] colorUsingColorSpaceName:NSDeviceRGBColorSpace];
			float r=[back redComponent],g=[back greenComponent],b=[back blueComponent];
			GLuint tex=make_grid_texture(r,g,b);

			int back_x=draw_x>0?draw_x:0;
			int back_y=draw_y>0?draw_y:0;
			int back_w=MIN(width,imgwidth);
			int back_h=MIN(height,imgheight);

			glDisable(GL_BLEND);
			glDisable(GL_TEXTURE_RECTANGLE_EXT);
			glEnable(GL_TEXTURE_2D);

			glPushMatrix();
			glTranslatef(back_x,back_y,0);

			glMatrixMode(GL_TEXTURE);
			glPushMatrix();
			GLfloat rot[]={1,1,0,0,1,-1,0,0,0,0,1,0,0,0,0,1};
			glLoadMatrixf(rot);
			glScalef(1.0/31.0,1.0/31.0,1.0/31.0);

			glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex2f(0,0);
			glTexCoord2f(back_w,0);
			glVertex2f(back_w,0);
			glTexCoord2f(back_w,back_h);
			glVertex2f(back_w,+back_h);
			glTexCoord2f(0,back_h);
			glVertex2f(0,back_h);
			glEnd();

			glPopMatrix();

			glMatrixMode(GL_MODELVIEW);
			glPopMatrix();

			glDeleteTextures(1,&tex);
		}

		[image draw:NSMakePoint(draw_x,draw_y) size:NSMakeSize(imgwidth,imgheight) bounds:[self bounds]];
	}

	if(drawresize)
	{
		GLuint tex=make_resize_texture();

		glEnable(GL_BLEND);
		glDisable(GL_TEXTURE_RECTANGLE_EXT);
		glEnable(GL_TEXTURE_2D);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

		float x1=width-16;
		float y1=height-16;
		float x2=width;
		float y2=height;

		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2f(x1,y1);
		glTexCoord2f(1,0);
		glVertex2f(x2,y1);
		glTexCoord2f(1,1);
		glVertex2f(x2,y2);
		glTexCoord2f(0,1);
		glVertex2f(x1,y2);
		glEnd();

		glDeleteTextures(1,&tex);
	}

	//[[self openGLContext] flushBuffer];
	glFlush();
}

-(void)reshape
{
	[[self openGLContext] makeCurrentContext];
	[[self openGLContext] update];

	NSPoint focus=[self focus];

	NSRect bounds=[self bounds];
	width=bounds.size.width;
	height=bounds.size.height;

	[self setFocus:focus];
}


-(BOOL)acceptsFirstResponder { return YES; }

-(BOOL)isOpaque { return YES; }

-(void)keyDown:(NSEvent *)event
{
	if(![[event characters] length])
	{
		[super keyDown:event];
		return;
	}

	unichar c=[[event characters] characterAtIndex:0];

	if(c==NSUpArrowFunctionKey) up=YES;
	else if(c==NSDownArrowFunctionKey) down=YES;
	else if(c==NSLeftArrowFunctionKey) left=YES;
	else if(c==NSRightArrowFunctionKey) right=YES;
	else [super keyDown:event];

	if(up||down||left||right) [self startScrolling];
}

-(void)keyUp:(NSEvent *)event
{
	if(![[event characters] length])
	{
		[super keyDown:event];
		return;
	}

	unichar c=[[event characters] characterAtIndex:0];

	if(c==NSUpArrowFunctionKey) up=NO;
	else if(c==NSDownArrowFunctionKey) down=NO;
	else if(c==NSLeftArrowFunctionKey) left=NO;
	else if(c==NSRightArrowFunctionKey) right=NO;
	else [super keyUp:event];

	if(!up&&!down&&!left&&!right) [self stopScrolling];
}

-(void)mouseDown:(NSEvent *)event
{
	[[self window] setAcceptsMouseMovedEvents:YES];
	[[self window] invalidateCursorRectsForView:self];
}

-(void)mouseUp:(NSEvent *)event
{
	[[self window] setAcceptsMouseMovedEvents:NO];
	[[self window] invalidateCursorRectsForView:self];
}

-(void)mouseDragged:(NSEvent *)event
{
	x-=[event deltaX];
	y-=[event deltaY];

	[self clampCoords];
	[self setNeedsDisplay:YES];
}

-(void)resetCursorRects
{
	if(width<imgwidth||height<imgheight)
	{
		NSCursor *cursor;
		if([[self window] acceptsMouseMovedEvents]) cursor=[NSCursor closedHandCursor];
		else cursor=[NSCursor openHandCursor];

		if(drawresize)
		{
			NSRect rect1=[self visibleRect];
			NSRect rect2=[self visibleRect];
			rect1.size.width-=15;
			rect2.size.height-=15;
			[self addCursorRect:rect1 cursor:cursor];
			[self addCursorRect:rect2 cursor:cursor];
		}
		else
		{
			[self addCursorRect:[self visibleRect] cursor:cursor];
		}
	}
}

/*-(NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
{
	return NSDragOperationGeneric;
}

-(BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
	NSPasteboard *pboard=[sender draggingPasteboard];
	NSArray *files=[pboard propertyListForType:NSFilenamesPboardType];

	[controller loadImage:[files objectAtIndex:0]];

    return YES;
}*/



-(void)setImage:(XeeImage *)img
{
	if(image!=img)
	{
		[image release];
		image=[img retain];
	}

	[self setNeedsDisplay:YES];
}

-(void)setImageSize:(NSSize)size
{
	NSPoint focus=[self focus];

	int oldwidth=imgwidth;
	int oldheight=imgheight;

	imgwidth=size.width;
	imgheight=size.height;

	focus.x*=(float)imgwidth/(float)oldwidth;
	focus.y*=(float)imgheight/(float)oldheight;

	[self setFocus:focus];

	[self setNeedsDisplay:YES];
}

-(void)setDrawResizeCorner:(BOOL)draw { drawresize=draw; }

-(void)startScrolling
{
	if(!scrolltimer)
	{
		prevtime=[self getTime];
		scrolltimer=[[NSTimer scheduledTimerWithTimeInterval:1.0/60.0 target:self selector:@selector(scroll:) userInfo:nil repeats:YES] retain];
	}
}

-(void)stopScrolling
{
	if(scrolltimer)
	{
		[scrolltimer invalidate];
		[scrolltimer release];
		scrolltimer=nil;
	}
}

-(void)scroll:(NSTimer *)timer
{
	double time=[self getTime];
	double dt=time-prevtime;
	prevtime=time;

	double mult=1.0;

	if(GetCurrentKeyModifiers()&((1<<shiftKeyBit)|(1<<rightShiftKeyBit))) mult=3.0;

	int delta=1500.0*mult*dt;

	if(left) x-=delta;
	if(right) x+=delta;
	if(up) y-=delta;
	if(down) y+=delta;

	[self clampCoords];
	[self setNeedsDisplay:YES];
}

-(double)getTime
{
	struct timeval tv;
	gettimeofday(&tv,0);
	return (double)tv.tv_sec+((double)tv.tv_usec)/1000000.0;
}



-(NSPoint)focus
{
	NSPoint focus;

	if(imgwidth<width) focus.x=width/2;
	else focus.x=x+width/2;

	if(imgheight<height) focus.y=height/2;
	else focus.y=y+height/2;

	return(focus);
}

-(void)setFocus:(NSPoint)focus
{
	x=focus.x-width/2;
	y=focus.y-height/2;

	[self clampCoords];
	[self setNeedsDisplay:YES];
}

-(void)clampCoords
{
	if(x>imgwidth-width) x=imgwidth-width;
	if(y>imgheight-height) y=imgheight-height;
	if(x<0) x=0;
	if(y<0) y=0;
}


-(BOOL)isFlipped { return YES; }

static const void *get_byte_pointer(void *bitmap) { return bitmap; }

-(void)copyGLtoQuartz
{
	int bytesperrow=width*4;
    void *bitmap=malloc(bytesperrow*height);

	[[self openGLContext] makeCurrentContext];

    glFinish(); // finish any pending OpenGL commands
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	//glPixelStorei(GL_PACK_SWAP_BYTES, 0);
	//glPixelStorei(GL_PACK_LSB_FIRST, 0);
	//glPixelStorei(GL_PACK_IMAGE_HEIGHT, 0);
	glPixelStorei(GL_PACK_ALIGNMENT,4); // force 4-byte alignment from RGBA framebuffer
	glPixelStorei(GL_PACK_ROW_LENGTH,0);
	glPixelStorei(GL_PACK_SKIP_PIXELS,0);
	glPixelStorei(GL_PACK_SKIP_ROWS,0);

	glReadPixels(0,0,width,height,GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,bitmap);
	glPopAttrib();

    [self lockFocus];

    CGDataProviderDirectAccessCallbacks callbacks={get_byte_pointer,NULL,NULL,NULL};
    CGDataProviderRef provider=CGDataProviderCreateDirectAccess(bitmap,bytesperrow*height,&callbacks);
    CGColorSpaceRef cs=CGColorSpaceCreateDeviceRGB();
    CGImageRef cgimage=CGImageCreate(width,height,8,32,bytesperrow,cs,kCGImageAlphaNoneSkipFirst,provider,NULL,NO,kCGRenderingIntentDefault);

    CGContextRef gc=[[NSGraphicsContext currentContext] graphicsPort];
    CGContextDrawImage(gc,CGRectMake(0,0,width,height),cgimage);

    CGImageRelease(cgimage);
    CGDataProviderRelease(provider);
    CGColorSpaceRelease(cs);
    free(bitmap);

	[self unlockFocus];

	[[self window] flushWindow];
}

-(void)defaultUpdateAction:(XeeImage *)img
{
	[self setNeedsDisplay:YES];
}

@end



GLuint make_grid_texture(float r,float g,float b)
{
	float r_low=r*0.9,g_low=g*0.9,b_low=b*0.9;
	float r_high=r_low+0.1,g_high=g_low+0.1,b_high=b_low+0.1;
	int r1=(int)(255.0*r_low),g1=(int)(255.0*g_low),b1=(int)(255.0*b_low);
	int r2=(int)(255.0*r_high),g2=(int)(255.0*g_high),b2=(int)(255.0*b_high);
	unsigned long col1=0xff000000|(r1<<16)|(g1<<8)|b1;
	unsigned long col2=0xff000000|(r2<<16)|(g2<<8)|b2;
	unsigned long data[]={col1,col2,col2,col1};

	GLuint tex;
	glGenTextures(1,&tex),
	glBindTexture(GL_TEXTURE_2D,tex);

	glPixelStorei(GL_UNPACK_ROW_LENGTH,2);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,GL_FALSE);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,2,2,0,GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,data);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

	return tex;
}

#define C0 0x00000000
#define C1 0x25000000
#define C2 0x8a222222
#define C3 0x8bd5d5d5

static unsigned long resize_data[256]=
{
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C1,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C1,C2,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C1,C2,C3,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C1,C2,C3,C0,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C1,C2,C3,C0,C1,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C1,C2,C3,C0,C1,C2,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C1,C2,C3,C0,C1,C2,C3,C0,
	C0,C0,C0,C0,C0,C0,C0,C1,C2,C3,C0,C1,C2,C3,C0,C0,
	C0,C0,C0,C0,C0,C0,C1,C2,C3,C0,C1,C2,C3,C0,C1,C0,
	C0,C0,C0,C0,C0,C1,C2,C3,C0,C1,C2,C3,C0,C1,C2,C0,
	C0,C0,C0,C0,C1,C2,C3,C0,C1,C2,C3,C0,C1,C2,C3,C0,
	C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,C0,
};

GLuint make_resize_texture()
{
	GLuint tex;
	glGenTextures(1,&tex),
	glBindTexture(GL_TEXTURE_2D,tex);

	glPixelStorei(GL_UNPACK_ROW_LENGTH,16);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,GL_TRUE);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,16,16,0,GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,resize_data);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

	return tex;
}